@@ -6,6 +6,8 @@ Localization utilities to find available language packs and packages with
 localization data.
 """
 
+from __future__ import annotations
+
 import gettext
 import importlib
 import json
@@ -15,7 +17,7 @@ import re
 import sys
 import traceback
 from functools import lru_cache
-from typing import Any, Dict, Optional, Pattern, Tuple
+from typing import Any, Pattern
 
 import babel
 from packaging.version import parse as parse_version
@@ -39,6 +41,9 @@ DEFAULT_DOMAIN = "jupyterlab"
 L10N_SCHEMA_NAME = "@jupyterlab/translation-extension:plugin"
 PY37_OR_LOWER = sys.version_info[:2] <= (3, 7)
 
+# Pseudo language locale for in-context translation
+PSEUDO_LANGUAGE = "ach_UG"
+
 _default_schema_context = "schema"
 _default_settings_context = "settings"
 _lab_i18n_config = "jupyter.lab.internationalization"
@@ -58,15 +63,15 @@ DEFAULT_SCHEMA_SELECTORS = {
 }
 
 
-@lru_cache()
-def _get_default_schema_selectors() -> Dict[Pattern, str]:
+@lru_cache
+def _get_default_schema_selectors() -> dict[Pattern, str]:
     return {
         re.compile("^/" + pattern + "$"): context
         for pattern, context in DEFAULT_SCHEMA_SELECTORS.items()
     }
 
 
-def _prepare_schema_patterns(schema: dict) -> Dict[Pattern, str]:
+def _prepare_schema_patterns(schema: dict) -> dict[Pattern, str]:
     return {
         **_get_default_schema_selectors(),
         **{
@@ -78,7 +83,7 @@ def _prepare_schema_patterns(schema: dict) -> Dict[Pattern, str]:
 
 # --- Private process helpers
 # ----------------------------------------------------------------------------
-def _get_installed_language_pack_locales():
+def _get_installed_language_pack_locales() -> tuple[dict[str, Any], str]:
     """
     Get available installed language pack locales.
 
@@ -100,7 +105,7 @@ def _get_installed_language_pack_locales():
     return data, message
 
 
-def _get_installed_package_locales():
+def _get_installed_package_locales() -> tuple[dict[str, Any], str]:
     """
     Get available installed packages containing locale information.
 
@@ -150,7 +155,9 @@ def is_valid_locale(locale_: str) -> bool:
     - Brazilian German: "de_BR"
     """
     # Add exception for Norwegian
-    if locale_ == "no_NO":
+    if locale_ in {
+        "no_NO",
+    }:
         return True
 
     valid = False
@@ -182,14 +189,19 @@ def get_display_name(locale_: str, display_locale: str = DEFAULT_LOCALE) -> str:
     """
     locale_ = locale_ if is_valid_locale(locale_) else DEFAULT_LOCALE
     display_locale = display_locale if is_valid_locale(display_locale) else DEFAULT_LOCALE
-    loc = babel.Locale.parse(locale_)
-    display_name = loc.get_display_name(display_locale)
+    try:
+        loc = babel.Locale.parse(locale_)
+        display_name = loc.get_display_name(display_locale)
+    except babel.UnknownLocaleError:
+        display_name = display_locale
     if display_name:
         display_name = display_name[0].upper() + display_name[1:]
-    return display_name  # type:ignore
+    return display_name  # type:ignore[return-value]
 
 
-def merge_locale_data(language_pack_locale_data, package_locale_data):
+def merge_locale_data(
+    language_pack_locale_data: dict[str, Any], package_locale_data: dict[str, Any]
+) -> dict[str, Any]:
     """
     Merge language pack data with locale data bundled in packages.
 
@@ -226,7 +238,7 @@ def merge_locale_data(language_pack_locale_data, package_locale_data):
     return result
 
 
-def get_installed_packages_locale(locale_: str) -> Tuple[dict, str]:
+def get_installed_packages_locale(locale_: str) -> tuple[dict, str]:
     """
     Get all jupyterlab extensions installed that contain locale data.
 
@@ -279,7 +291,7 @@ def get_installed_packages_locale(locale_: str) -> Tuple[dict, str]:
 
 # --- API
 # ----------------------------------------------------------------------------
-def get_language_packs(display_locale: str = DEFAULT_LOCALE) -> Tuple[dict, str]:
+def get_language_packs(display_locale: str = DEFAULT_LOCALE) -> tuple[dict, str]:
     """
     Return the available language packs installed in the system.
 
@@ -309,21 +321,38 @@ def get_language_packs(display_locale: str = DEFAULT_LOCALE) -> Tuple[dict, str]
             else:
                 invalid_locales.append(locale_)
 
-        display_locale = display_locale if display_locale in valid_locales else DEFAULT_LOCALE
+        display_locale_ = display_locale if display_locale in valid_locales else DEFAULT_LOCALE
         locales = {
             DEFAULT_LOCALE: {
-                "displayName": get_display_name(DEFAULT_LOCALE, display_locale),
+                "displayName": (
+                    get_display_name(DEFAULT_LOCALE, display_locale_)
+                    if display_locale != PSEUDO_LANGUAGE
+                    else "Default"
+                ),
                 "nativeName": get_display_name(DEFAULT_LOCALE, DEFAULT_LOCALE),
             }
         }
         for locale_ in valid_locales:
             locales[locale_] = {
-                "displayName": get_display_name(locale_, display_locale),
+                "displayName": get_display_name(locale_, display_locale_),
                 "nativeName": get_display_name(locale_, locale_),
             }
 
         if invalid_locales:
-            messages.append(f"The following locales are invalid: {invalid_locales}!")
+            if PSEUDO_LANGUAGE in invalid_locales:
+                invalid_locales.remove(PSEUDO_LANGUAGE)
+                locales[PSEUDO_LANGUAGE] = {
+                    "displayName": "Pseudo-language",
+                    # Trick to ensure the proper language is selected in the language menu
+                    "nativeName": (
+                        "to translate the UI"
+                        if display_locale != PSEUDO_LANGUAGE
+                        else "Pseudo-language"
+                    ),
+                }
+            # Check again as the pseudo-language was maybe the only invalid locale
+            if invalid_locales:
+                messages.append(f"The following locales are invalid: {invalid_locales}!")
 
     return locales, "\n".join(messages)
 
@@ -348,7 +377,11 @@ def get_language_pack(locale_: str) -> tuple:
     found_packages_locales, message = get_installed_packages_locale(locale_)
     locale_data = {}
     messages = message.split("\n")
-    if not message and is_valid_locale(locale_) and locale_ in found_locales:
+    if (
+        not message
+        and (locale_ == PSEUDO_LANGUAGE or is_valid_locale(locale_))
+        and locale_ in found_locales
+    ):
         path = found_locales[locale_]
         for root, __, files in os.walk(path, topdown=False):
             for name in files:
@@ -407,9 +440,9 @@ class TranslationBundle:
             language_pack_module = f"jupyterlab_language_pack_{locale_}"
             try:
                 mod = importlib.import_module(language_pack_module)
-                assert mod.__file__ is not None  # noqa
+                assert mod.__file__ is not None
                 localedir = os.path.join(os.path.dirname(mod.__file__), LOCALE_DIR)
-            except Exception:  # noqa S110
+            except Exception:  # noqa: S110
                 # no-op
                 pass
 
@@ -585,12 +618,12 @@ class TranslationBundle:
         return self.npgettext(msgctxt, msgid, msgid_plural, n)
 
 
-class translator:  # noqa
+class translator:
     """
     Translations manager.
     """
 
-    _TRANSLATORS: Dict[str, TranslationBundle] = {}
+    _TRANSLATORS: dict[str, TranslationBundle] = {}
     _LOCALE = SYS_LOCALE
 
     @staticmethod
@@ -659,7 +692,7 @@ class translator:  # noqa
         translations: Any,
         schema: dict,
         prefix: str = "",
-        to_translate: Optional[Dict[Pattern, str]] = None,
+        to_translate: dict[Pattern, str] | None = None,
     ) -> None:
         """Translate a schema in-place."""
         if to_translate is None:
@@ -670,7 +703,7 @@ class translator:  # noqa
 
             if isinstance(value, str):
                 matched = False
-                for pattern, context in to_translate.items():  # noqa
+                for pattern, context in to_translate.items():  # noqa: B007
                     if pattern.fullmatch(path):
                         matched = True
                         break
@@ -695,7 +728,7 @@ class translator:  # noqa
                     )
 
     @staticmethod
-    def translate_schema(schema: Dict) -> Dict:
+    def translate_schema(schema: dict) -> dict:
         """Translate a schema.
 
         Parameters
